#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "../include/string.h"

#include "obj.h"
#include "sar.h"
#include "messages.h"


void SARMessageAdd(sar_scene_struct *scene, const char *message);
void SARMessageClearAll(sar_scene_struct *scene);


/*
 *	Adds the specified message to the list in the scene.
 *	The new message will be inserted at the last index position.
 *
 *	The message pointer array on the scene will not change in size.
 */
void SARMessageAdd(sar_scene_struct *scene, const char *message)
{
	int i, n;


	if(scene == NULL)
	    return;

	/* No message line pointers allocated? */
	if(scene->total_messages <= 0)
	    return;

	/* Free oldest message. */
	free(scene->message[0]);

	/* Get new position to insert new message. */
	n = scene->total_messages - 1;

	/* Shift message pointers. */
        for(i = 0; i < n; i++)
            scene->message[i] = scene->message[i + 1];

	/* Copy new message to pointer array. */
	scene->message[n] = ((message == NULL) ? NULL : strdup(message));

	/* Update the time the message should be shown untill. */
	scene->message_display_untill = cur_millitime +
	    SAR_MESSAGE_SHOW_INT;

	return;
}

/*
 *	Deletes all messages on the message pointer array on the scene
 *	structure and resets all pointers to NULL.
 *
 *	The size of the message pointer array will not be changed.
 *
 *	The message_display_untill member on the scene will be reset
 *	to 0.
 */
void SARMessageClearAll(sar_scene_struct *scene)
{
	int i;

	if(scene == NULL)
	    return;

	/* Reset the time to keep displaying messages to 0 so messages
	 * shouldn't be displayed anymore if they were being displayed.
	 */
	scene->message_display_untill = 0;

	/* Delete each message and set pointer to NULL. */
	for(i = 0; i < scene->total_messages; i++)
	{
	    free(scene->message[i]);
	    scene->message[i] = NULL;
	}

	return;
}
