#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "gw.h"

#include "messages.h"
#include "sarreality.h"
#include "obj.h"
#include "gctl.h"
#include "simutils.h"
#include "explosion.h"
#include "textinput.h"
#include "cmd.h"

#include "sar.h"


static void SARKeyCameraRefCockpit(
        sar_core_struct *core_ptr,
        gw_display_struct *display,
        sar_scene_struct *scene,
        Boolean state
);
static void SARKeyCameraRefHoist( 
        sar_core_struct *core_ptr, 
        gw_display_struct *display,
        sar_scene_struct *scene,
        Boolean state
);
static void SARKeyCameraRefMap(
        sar_core_struct *core_ptr,
        gw_display_struct *display,
        sar_scene_struct *scene,
        Boolean state
);
static void SARKeyCameraRefSpot(
        sar_core_struct *core_ptr,
        gw_display_struct *display,
        sar_scene_struct *scene,
        Boolean state
);
static void SARKeyCameraRefTower(
        sar_core_struct *core_ptr,
        gw_display_struct *display,
        sar_scene_struct *scene,
        Boolean state
);
static void SARKeyCommand(
        sar_core_struct *core_ptr,
        gw_display_struct *display,
        sar_scene_struct *scene,
        Boolean state
);
static void SARKeyEscape(
        sar_core_struct *core_ptr,
        gw_display_struct *display,
        sar_scene_struct *scene,
        Boolean state
);

static void SARKeyGraphicsAtmosphere(
        sar_core_struct *core_ptr,
        gw_display_struct *display,
        sar_scene_struct *scene,   
        Boolean state
);
static void SARKeyGraphicsTexturedClouds(
        sar_core_struct *core_ptr,
        gw_display_struct *display,
        sar_scene_struct *scene,
        Boolean state
);
static void SARKeyGraphicsTexturedGround(
        sar_core_struct *core_ptr,
        gw_display_struct *display,
        sar_scene_struct *scene,
        Boolean state
);
static void SARKeyGraphicsTexturedObjects(
        sar_core_struct *core_ptr,
        gw_display_struct *display,
        sar_scene_struct *scene,   
        Boolean state
);
static void SARKeyHelpDisplayToggle(
        sar_core_struct *core_ptr,
        gw_display_struct *display,
        sar_scene_struct *scene,
        Boolean state 
);
static void SARKeyPrintScore(
	sar_core_struct *core_ptr,
	gw_display_struct *display,
	sar_scene_struct *scene,
	Boolean state
);
static void SARKeyTimeCompression( 
        sar_core_struct *core_ptr, 
        gw_display_struct *display,
        sar_scene_struct *scene,
        Boolean state
);
static void SARKeyViewNormalize(
        sar_core_struct *core_ptr,
        gw_display_struct *display,
        sar_scene_struct *scene,
        Boolean state
);

void SARKey(
	sar_core_struct *core_ptr,
	int c, Boolean state, long t
);


#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))


/*
 *	Switch camera referance on scene structure to cockpit.
 */
static void SARKeyCameraRefCockpit(
        sar_core_struct *core_ptr, 
        gw_display_struct *display,
        sar_scene_struct *scene,
        Boolean state
)
{
	if((scene == NULL) || !state)
	    return;

	scene->camera_ref = SAR_CAMERA_REF_COCKPIT;
	scene->camera_target = scene->player_obj_num;

	return;
}

/*                     
 *      Switch camera referance on scene structure to cockpit.
 */
static void SARKeyCameraRefHoist(
        sar_core_struct *core_ptr, 
        gw_display_struct *display,
        sar_scene_struct *scene,
        Boolean state
)
{
	if((core_ptr == NULL) || (scene == NULL) || !state)
	    return;

	scene->camera_ref = SAR_CAMERA_REF_HOIST;
	scene->camera_target = scene->player_obj_num;

	return;
}

/*
 *      Switch camera referance on scene structure to map.
 */
static void SARKeyCameraRefMap(
        sar_core_struct *core_ptr,
        gw_display_struct *display,
        sar_scene_struct *scene,
        Boolean state
)
{
	sar_object_struct *obj_ptr;

	if((core_ptr == NULL) || (scene == NULL) || !state)
	    return;

	scene->camera_ref = SAR_CAMERA_REF_MAP;
	scene->camera_target = -1;

	obj_ptr = scene->player_obj_ptr;
	if(obj_ptr != NULL)
	{
	    sar_position_struct *pos_src = &obj_ptr->pos;
	    sar_position_struct *pos_tar = &scene->camera_pos;

	    /* Set initial camera position at location of object. */
	    pos_tar->x = pos_src->x;
	    pos_tar->y = pos_src->y;

	    if(pos_tar->z < 1000.0)
		pos_tar->z = 1000.0;
	}

	return;
}


/*
 *      Switch camera referance on scene structure to spot.
 *
 *	If the camera referance is already set to spot then the `next'
 *	object will be set as the camera target.
 */
static void SARKeyCameraRefSpot(
        sar_core_struct *core_ptr,
        gw_display_struct *display, 
        sar_scene_struct *scene,  
        Boolean state
)
{
        if((core_ptr == NULL) || (scene == NULL) || !state)
            return;

	/* Previous camera referance is spot? */
	if(scene->camera_ref == SAR_CAMERA_REF_SPOT)
	{
	    /* Camera referance was already spot, so cycle camera target
	     * to the `next' object.
	     */
	    int i, n, type;
	    sar_object_struct *obj_ptr;

	    /* Go through objects list twice. */
	    for(n = 0; n < 2; n++)
	    {
		/* Go through objects list, starting from the
		 * index value as the current value for the camera
		 * target + 1.
		 */
		for(i = MAX(scene->camera_target + 1, 0);
		    i < core_ptr->total_objects;
		    i++
		)
		{
		    obj_ptr = core_ptr->object[i];
		    if(obj_ptr == NULL)
			continue;

		    /* Get object type. */
		    type = obj_ptr->type;

		    /* Skip following object types. */
		    if((type == SAR_OBJ_TYPE_STATIC) ||
		       (type == SAR_OBJ_TYPE_AUTOMOBILE) ||
		       (type == SAR_OBJ_TYPE_WATERCRAFT) ||
		       (type == SAR_OBJ_TYPE_AIRCRAFT) ||
/*
		       (type == SAR_OBJ_TYPE_RUNWAY) ||
		       (type == SAR_OBJ_TYPE_HELIPAD) ||
 */
		       (type == SAR_OBJ_TYPE_HUMAN) ||
		       (type == SAR_OBJ_TYPE_PREMODELED)
		    )
			break;
		}
		if(i < core_ptr->total_objects)
		{
		    scene->camera_target = i;
		    break;
		}
		scene->camera_target = -1;
	    }
	}
	else
	{
	    scene->camera_ref = SAR_CAMERA_REF_SPOT;
	    scene->camera_target = scene->player_obj_num;
	}

        return;
}

/*
 *      Switch camera referance on scene structure to tower.
 *
 *	The position of the tower will be calculated by a call to
 *	SARSimSetFlyByPosition().
 */
static void SARKeyCameraRefTower(
        sar_core_struct *core_ptr,
        gw_display_struct *display,
        sar_scene_struct *scene,
        Boolean state
)
{
        if((core_ptr == NULL) || (scene == NULL) || !state)
            return;

	scene->camera_ref = SAR_CAMERA_REF_TOWER;
	scene->camera_target = scene->player_obj_num;

	SARSimSetFlyByPosition(
	    scene,
	    &core_ptr->object, &core_ptr->total_objects,
	    scene->player_obj_ptr,
	    &scene->camera_pos
	);

	return;
}

/*
 *	Maps the command prompt, future key events sent to SARKey()
 *	will then be forwarded to SARTextInputHandleKey() untill
 *	the command argument is typed in and processed or aborted.
 */
static void SARKeyCommand(
        sar_core_struct *core_ptr,
        gw_display_struct *display,
        sar_scene_struct *scene,
        Boolean state
)
{
	if((core_ptr == NULL) || !state)
	    return;

	SARTextInputMap(
	    core_ptr,
	    "Command",
	    SARCommandTextInputCB
	);

	return;
}


/*
 *	Handles escape key depending on the situation which is checked
 *	in a specific order. The conditions are as follows:
 *
 *	1. Check is help screen is displayed, if so then hide it.
 *
 *	2. Check if any messages are currently being displayed by the
 *	   scene.
 *
 *	3. Check if a mission is in progress, and if so map the prompt
 *	   for confirmation.
 *
 *	4. All else prompt to end simulation.
 */
static void SARKeyEscape(
        sar_core_struct *core_ptr,
        gw_display_struct *display,
        sar_scene_struct *scene,
        Boolean state
)
{
        if((core_ptr == NULL) || !state)
            return;

	/* Displaying help? */
	if(core_ptr->display_help)
	{
	    core_ptr->display_help = False;
	}
	/* Scene displaying message? */
	else if((scene == NULL) ? 0 :
	        (scene->message_display_untill > 0)
        )
        {
	    /* Reset message display untill time to 0 and delete
	     * all messages on the scene. This will not change
	     * the size of the message pointer array.
	     */
	    SARMessageClearAll(scene);
	}
	/* Prompt to end mission? */
	else if(core_ptr->mission != NULL)
	{
	    /* Same as end simulation, just prompt text different. */
	    SARTextInputMap(
		core_ptr,
		"Are you sure you want to abort the mission?",
		SARTextInputCBQuitSimulation
	    );
	}
	else
	{
	    SARTextInputMap(
		core_ptr,
		"Are you sure you want to quit simulation?",
		SARTextInputCBQuitSimulation
	    );
	}

	return;
}


/*
 *      Toggles atmosphere in the global options structure on/off.
 */
static void SARKeyGraphicsAtmosphere(
        sar_core_struct *core_ptr,
        gw_display_struct *display,
        sar_scene_struct *scene,
        Boolean state
)
{
        const char *message = NULL;

        if((core_ptr == NULL) || (scene == NULL) || !state)
            return;

        if(option.atmosphere)
        {
            option.atmosphere = False;
            message = "Atmosphere: Off";
        }
        else
        {
            option.atmosphere = True;
            message = "Atmosphere: On";
        }

        SARMessageAdd(scene, message);

        return;
}

/*
 *	Toggles textured cloud layers in the global options structure
 *	on/off.
 *
 *	Note that this basically specifies if cloud layers should be
 *	drawn since cloud layers are always textured. If camera is
 *	above highest cloud layer then the highest cloud layer will
 *	always be drawn.
 */
static void SARKeyGraphicsTexturedClouds(
        sar_core_struct *core_ptr, 
        gw_display_struct *display,
        sar_scene_struct *scene,
        Boolean state
)
{
        const char *message = NULL;

        if((core_ptr == NULL) || (scene == NULL) || !state)
            return;

        if(option.textured_clouds)
        {
            option.textured_clouds = False;
            message = "Textured Clouds: Off";
        }
        else
        {
            option.textured_clouds = True;
            message = "Textured Clouds: On";
        }

        SARMessageAdd(scene, message);

        return;
}

/*
 *	Toggles textured ground in the global options structure on/off.
 */
static void SARKeyGraphicsTexturedGround(
        sar_core_struct *core_ptr,
        gw_display_struct *display,
        sar_scene_struct *scene,
        Boolean state
)
{
	const char *message = NULL;

        if((core_ptr == NULL) || (scene == NULL) || !state)
            return; 

	if(option.textured_ground)
	{
	    option.textured_ground = False;
	    message = "Ground Texture: Off";
	}
	else
	{
	    option.textured_ground = True;
	    message = "Ground Texture: On";
	}

	SARMessageAdd(scene, message);

	return;
}

/*
 *      Toggles textured objects in the global options structure on/off.
 */
static void SARKeyGraphicsTexturedObjects(
        sar_core_struct *core_ptr,
        gw_display_struct *display,
        sar_scene_struct *scene,
        Boolean state
)
{
        const char *message = NULL;

        if((core_ptr == NULL) || (scene == NULL) || !state)
            return;

        if(option.textured_objects)
        {
            option.textured_objects = False;
            message = "Object Texture: Off";
        }
        else
        {
            option.textured_objects = True;
            message = "Object Texture: On";
        }

        SARMessageAdd(scene, message);

        return;
}


/*
 *	Toggles the value display_help on the core structure on/off.
 *
 *	This specifies if the help screen should be drawn or not.
 */
static void SARKeyHelpDisplayToggle(
        sar_core_struct *core_ptr, 
        gw_display_struct *display,
        sar_scene_struct *scene,
        Boolean state
)
{
        if((core_ptr == NULL) || !state)
            return;

	core_ptr->display_help = !core_ptr->display_help;

	return;
}

/*
 *	Prints score, mission status, and passengers on board player
 *	aircraft.
 */
static void SARKeyPrintScore(
        sar_core_struct *core_ptr,
        gw_display_struct *display,
        sar_scene_struct *scene, 
        Boolean state  
)
{
	sar_object_struct *obj_ptr;

	if((core_ptr == NULL) || (display == NULL) || (scene == NULL) || !state)
	    return;

	/* Get player object pointer from scene structure. */
	obj_ptr = scene->player_obj_ptr;
	if(obj_ptr != NULL)
	{
	    if(display->shift_key_state)
	    {
		/* Debug case, print player position stats to stdout. */
		sar_direction_struct *dir = &obj_ptr->dir;
		sar_position_struct *pos = &obj_ptr->pos;

		/* Print position (for debugging). */
		printf(
 "%i:%2i Player: (hpb) %.0f %.0f %.0f  (xyz) %.2f %.2f %.2f(%.2f feet)\n",
		    (int)(scene->tod / 3600),
		    (int)((int)((int)scene->tod / 60) % 60),
		    SFMRadiansToDegrees(dir->heading),
		    SFMRadiansToDegrees(dir->pitch),
		    SFMRadiansToDegrees(dir->bank),
		    pos->x, pos->y, pos->z,
		    SFMMetersToFeet(pos->z)
		);
	    }
	    else
	    {
		/* Print score, mission status, and occupants on
		 * the given player object obj_ptr. If core_ptr->mission
		 * is NULL then only occupants will be printed.
		 */
		SARMissionPrintStats(
		    core_ptr, scene, core_ptr->mission,
		    obj_ptr
		);
	    }
	}

	return;
}

/*
 *	Adjusts the global time compression.
 */
static void SARKeyTimeCompression(
        sar_core_struct *core_ptr, 
        gw_display_struct *display,
        sar_scene_struct *scene,
        Boolean state
)
{
	char text[256];

        if((core_ptr == NULL) || (scene == NULL) || !state)
            return;

	/* Check shift modifier. */
	if(display->shift_key_state)
	{
	    /* Shift is held, so increase time compression. */
	    if(time_compression >= 1.0)
	    {
		time_compression += 1.0;
	    }
	    else
	    {
		time_compression += 0.25;
		if(time_compression > 1.0)
		    time_compression = 1.0;
	    }
	}
	else
	{
	    /* Shift is held, so decrease time compression. */
	    if(time_compression > 1.0)
	    {
		time_compression -= 1.0;
		if(time_compression < 1.0)
		    time_compression = 1.0;
	    }
	    else
	    {
		time_compression -= 0.25;
	    }
	}

        if(time_compression < 0.25)
	    time_compression = 0.25;
	else if(time_compression > 8.0)
	    time_compression = 8.0;

	sprintf(
	    text,
	    "Time Compression %.0f%%",
	    time_compression * 100
	);
	SARMessageAdd(scene, text);

	return;
}

/*
 *	Normalizes the zoom and position of the view with respect to
 *	the current camera_ref on the scene.
 */
static void SARKeyViewNormalize(
        sar_core_struct *core_ptr,
        gw_display_struct *display,
        sar_scene_struct *scene,
        Boolean state
)
{
        sar_object_struct *obj_ptr;

        if((core_ptr == NULL) || (display == NULL) || (scene == NULL) || !state)
            return;

	/* Get player object pointer from scene structure. */
	obj_ptr = scene->player_obj_ptr;

	/* Handle by camera referance specified on the scene structure. */
	switch(scene->camera_ref)
	{
	  case SAR_CAMERA_REF_HOIST:
	    if(obj_ptr != NULL)
	    {
		double contact_radius = SARSimGetFlatContactRadius(obj_ptr);
		sar_direction_struct *dir = &scene->camera_hoist_dir;

		scene->camera_hoist_dist = MAX(contact_radius, 10.0);
		dir->heading = (1.0 * PI);
		dir->pitch = (1.95 * PI);
		dir->bank = (0.0 * PI);
	    }
	    break;

	  case SAR_CAMERA_REF_MAP:
	    if(obj_ptr != NULL)
	    {
		sar_position_struct *pos_src = &obj_ptr->pos;
		sar_position_struct *pos_tar = &scene->camera_pos;

		/* Set camera position at location of object
		 * and keey the camera position atleast 1000.0 meters up.
		 */
		pos_tar->x = pos_src->x;
		pos_tar->y = pos_src->y;

		if(pos_tar->z < 1000.0)
		    pos_tar->z = 1000.0;
	    }
	    break;

          case SAR_CAMERA_REF_TOWER:
	    if(obj_ptr != NULL)
	    {
		SARSimSetFlyByPosition(
		    scene,
		    &core_ptr->object, &core_ptr->total_objects,
		    obj_ptr,
		    &scene->camera_pos
		);
	    }
            break;

          case SAR_CAMERA_REF_SPOT:
            if(obj_ptr != NULL)
	    {
		double contact_radius = SARSimGetFlatContactRadius(obj_ptr);
                sar_direction_struct *dir = &scene->camera_spot_dir;

		scene->camera_spot_dist = MAX(contact_radius, 10.0);
		dir->heading = (1.0 * PI);
		dir->pitch = (1.95 * PI);
		dir->bank = (0.0 * PI);
	    }
	    break;

	  case SAR_CAMERA_REF_COCKPIT:
	    if(1)
            {
		sar_direction_struct *dir = &scene->camera_cockpit_dir;

		dir->heading = (0.0 * PI);
		dir->pitch = (0.0 * PI);
		dir->bank = (0.0 * PI);
	    }
	    break;
	}

	return;
}


/*
 *	SAR keyboard event handler. Front end for handling all in game
 *	keyboard events.
 *
 *	If the in game text input prompt is mapped then the key event
 *	will be forwarded to the prompt input handler.
 *
 *	If key event key code c does not match any key operation to be
 *	handled then the keyboard event will be forwarde to
 *	GCtlHandleKey().
 */
void SARKey(
	sar_core_struct *core_ptr,
	int c, Boolean state, long t
)
{
	gw_display_struct *display;
	sar_scene_struct *scene;


	if(core_ptr == NULL)
	    return;

	display = core_ptr->display;
	scene = core_ptr->scene;

#define DO_HAS_NO_AUTOREPEAT	\
{ \
 GWKeyboardAutoRepeat(display, (Boolean)!state); \
}

	/* Handle key. */
	switch(c)
	{
/* Debug value adjustment. */
case 'x': case 'X':
 if((scene != NULL) && (display != NULL) && state)
 {
  debug_value += ((display->shift_key_state) ? 0.25 : -0.25);
fprintf(stderr, "Debug value set to %.2f\n", debug_value);
 }
 break;

	  case 's': case 'S':	/* Print score. */
	    DO_HAS_NO_AUTOREPEAT
	    SARKeyPrintScore(core_ptr, display, scene, state);
	    break;

	  case GWKeyBackSpace:	/* Re-center/normalize view. */
	    DO_HAS_NO_AUTOREPEAT
            SARKeyViewNormalize(core_ptr, display, scene, state);
	    break;

	  case GWKeyF1:		/* Help display toggle. */
	    DO_HAS_NO_AUTOREPEAT
            SARKeyHelpDisplayToggle(core_ptr, display, scene, state);
	    break;

          case GWKeyF2:		/* Switch camera ref to cockpit. */
	    DO_HAS_NO_AUTOREPEAT
	    SARKeyCameraRefCockpit(core_ptr, display, scene, state);
            break;

          case GWKeyF3:		/* Switch camera ref to spot. */
	    DO_HAS_NO_AUTOREPEAT
	    SARKeyCameraRefSpot(core_ptr, display, scene, state);
            break;

          case GWKeyF4:		/* Switch camera ref to tower. */
	    DO_HAS_NO_AUTOREPEAT
            SARKeyCameraRefTower(core_ptr, display, scene, state);
	    break;

          case GWKeyF5:		/* Switch camera ref to hoist's rescue basket. */
            DO_HAS_NO_AUTOREPEAT
	    SARKeyCameraRefHoist(core_ptr, display, scene, state);
	    break;

	  case 'm': case 'M':	/* Switch camera ref to map. */
	    DO_HAS_NO_AUTOREPEAT
	    SARKeyCameraRefMap(core_ptr, display, scene, state);
	    break;

	  case GWKeyF9:		/* Toggle textured ground. */
	    DO_HAS_NO_AUTOREPEAT
	    SARKeyGraphicsTexturedGround(core_ptr, display, scene, state);
            break;

          case GWKeyF10:	/* Toggle atmosphere. */
	    DO_HAS_NO_AUTOREPEAT
            SARKeyGraphicsAtmosphere(core_ptr, display, scene, state);
	    break;

          case GWKeyF11:	/* Toggle textured objects. */
	    DO_HAS_NO_AUTOREPEAT
	    SARKeyGraphicsTexturedObjects(core_ptr, display, scene, state);
	    break;

          case GWKeyF12:	/* Toggle textured clouds. */
	    DO_HAS_NO_AUTOREPEAT
	    SARKeyGraphicsTexturedClouds(core_ptr, display, scene, state);
	    break;

	  case 0x1b:		/* Escape. */
	    DO_HAS_NO_AUTOREPEAT
	    SARKeyEscape(core_ptr, display, scene, state);
            break;

	  case '/':		/* Command prompt map. */
	    DO_HAS_NO_AUTOREPEAT
	    SARKeyCommand(core_ptr, display, scene, state);
            break;

          case 'z': case 'Z':	/* Adjust time compression. */
            DO_HAS_NO_AUTOREPEAT
            SARKeyTimeCompression(core_ptr, display, scene, state);
            break;

	  /* Raise/lower landing gears on player object. */
	  case 'g': case 'G':
	    DO_HAS_NO_AUTOREPEAT
            if((scene != NULL) && state)
	    {
                sar_object_struct *obj_ptr;


                obj_ptr = scene->player_obj_ptr;
		if(obj_ptr != NULL)
		{
		    int lgear_state;


		    lgear_state = SARSimGetLandingGearState(obj_ptr);

		    /* Raise/lower landing gear. */
		    SARSimOpLandingGear(
			scene,
			&core_ptr->object, &core_ptr->total_objects,
			obj_ptr,
			!lgear_state
		    );
		}
	    }
	    break;

	  /* Turn lights on/off on player object. */
	  case 'l': case 'L':
	    if((scene != NULL) && state)
            {
                sar_object_struct *obj_ptr;

                obj_ptr = scene->player_obj_ptr;
                if(obj_ptr != NULL)
                {  
                    int state;

                    state = SARSimGetLightsState(obj_ptr);
                    SARSimOpLights(
                        scene,
                        &core_ptr->object, &core_ptr->total_objects,
                        obj_ptr,
                        !state
                    );
		    SFMSimOpAttenuate(
                        scene,
                        &core_ptr->object, &core_ptr->total_objects,
                        obj_ptr,
                        !state
                    );
                }
	    }
	    break;

          /* Turn strobe lights on/off on player object. */
          case 'o': case 'O':
            if((scene != NULL) && state)
            {
                sar_object_struct *obj_ptr;

                obj_ptr = scene->player_obj_ptr;
                if(obj_ptr != NULL)
                {
                    int state;

                    state = SARSimGetStrobesState(obj_ptr);
                    SARSimOpStrobes(
                        scene,
                        &core_ptr->object, &core_ptr->total_objects,
                        obj_ptr,
                        !state
                    );
                }
            }    
            break;

	  /* Adjust text color. */
	  case 'h': case 'H':
	    if((scene != NULL) && (display != NULL) && state)
	    {
		sar_color_struct *color;
		double g, b;	/* Gamma and base coeff. */


		/* Adjust hud color. */
		color = &option.hud_color;
		if(display->shift_key_state)
		{
		    /* Darken. */

		    if(color->r > 0)
		    {
			g = MAX(color->r - 0.2, 0);
			b = 1.0;
		    }
		    else
		    {
			g = 0;
			b = MAX(color->g - 0.2, 0);
		    }
		}
		else
		{
		    /* Lighten. */

		    if(color->g < 1)
		    {
			g = 0;
			b = MIN(color->g + 0.2, 1);
                    }
		    else
		    {
			g = MIN(color->r + 0.2, 1);
			b = 1;
		    }
		}
		color->a = 1.0;
		color->r = g;
		color->g = b;
		color->b = g;


	        /* Adjust message text color. */
                color = &option.message_color;
		if(display->shift_key_state)
		    g = MAX(color->r - 0.1, 0);
		else
		    g = MIN(color->r + 0.1, 1);

		color->a = 1.0,
		color->r = g;
		color->g = g;
		color->b = g;
	    }
	    break;

	  /* Select next intercept way point. */
	  case 'w': case 'W':
	    DO_HAS_NO_AUTOREPEAT
	    if((scene != NULL) && (display != NULL) && state)
            {
                sar_object_struct *obj_ptr;


                obj_ptr = scene->player_obj_ptr;
                if(obj_ptr != NULL)
                {
		    char text[256];
                    int type = obj_ptr->type;

		    *text = '\0';
                    if(type == SAR_OBJ_TYPE_AIRCRAFT)
                    {
                        sar_object_aircraft_struct *obj_aircraft_ptr =
                            obj_ptr->data;
                        if(obj_aircraft_ptr == NULL)
			    break;

			/* Change intercept number. */
			if(display->shift_key_state)
			    obj_aircraft_ptr->cur_intercept--;
			else
			    obj_aircraft_ptr->cur_intercept++;

			/* Cycle. */
			if(obj_aircraft_ptr->cur_intercept >=
			    obj_aircraft_ptr->total_intercepts
			)
			    obj_aircraft_ptr->cur_intercept = 0;
			else if(obj_aircraft_ptr->cur_intercept < 0)
			    obj_aircraft_ptr->cur_intercept =
				obj_aircraft_ptr->total_intercepts - 1;

			if(obj_aircraft_ptr->cur_intercept < 0)
			    obj_aircraft_ptr->cur_intercept = 0;

			if(obj_aircraft_ptr->total_intercepts > 0)
			    sprintf(text, "Selected waypoint %i.",
				obj_aircraft_ptr->cur_intercept
			    );
			else
			    sprintf(text, "No waypoints set.");
		    }

		    if(*text != '\0')
			SARMessageAdd(scene, text);
		}
	    }
	    break;

	  /* Engine off/init. */
	  case 'e': case 'E':
	    DO_HAS_NO_AUTOREPEAT
	    if((scene != NULL) && state)
	    {
                sar_object_struct *obj_ptr;   


                obj_ptr = scene->player_obj_ptr;
                if(obj_ptr != NULL)
		{
		    int engine_state;


		    engine_state = SARSimGetEngineState(obj_ptr);
		    if((engine_state == SAR_ENGINE_STATE_ON) ||
                       (engine_state == SAR_ENGINE_STATE_INIT)
		    )
			SARSimOpEngine(
			    scene,
			    &core_ptr->object, &core_ptr->total_objects,
			    obj_ptr,
			    SAR_ENGINE_STATE_OFF
			);
		    else
			SARSimOpEngine(
			    scene,
			    &core_ptr->object, &core_ptr->total_objects,
			    obj_ptr,
			    SAR_ENGINE_STATE_ON
			);
		    break;
		}
	    }
	    break;

	  /* Send message. */
	  case '\'':
	    if((scene != NULL) && (display != NULL) && state)
            {
		SARTextInputMap(
		    core_ptr,
		    "Message",
		    SARTextInputCBSendMessage
		);
	    }
	    break;

	  /* Adjust time of day. */
          case 't': case 'T':
            if((scene != NULL) && (display != NULL) && state)
            {
		char text[80];


		/* Change main scene light color for now. */
                if(display->shift_key_state)
		    scene->tod -= (0.25 * 3600);
		else
		    scene->tod += (0.25 * 3600);

		sprintf(
		    text,
		    "Time of day %i:%2.0i",
		    (int)(scene->tod / 3600),
		    (int)((int)((int)scene->tod / 60) % 60)
		);
                SARMessageAdd(scene, text);
	    }
	    break;

	  /* Open/close rescue door on player object. */
	  case 'd': case 'D':
	    DO_HAS_NO_AUTOREPEAT
            if((scene != NULL) && state)
            {
		/* Adjust flight physics dificulty. */
		if(display->ctrl_key_state)
		{
		    char *strptr = "Unsupported flight physics mode";

		    switch(option.flight_physics)
		    {
		      case FLIGHT_PHYSICS_EASY:
			option.flight_physics = FLIGHT_PHYSICS_MODERATE;
			strptr = "Flight Physics: Moderate";
			break;

		      case FLIGHT_PHYSICS_MODERATE:
                        option.flight_physics = FLIGHT_PHYSICS_REALISTIC;
                        strptr = "Flight Physics: Realistic";
			break;

		      default:
			option.flight_physics = FLIGHT_PHYSICS_EASY;
			strptr = "Flight Physics: Easy";
			break;
		    }
                    SARMessageAdd(scene, strptr);
		}
		/* Open/close door. */
		else
		{
		    int type;
                    sar_object_struct *obj_ptr;
		    sar_obj_door_struct *door_ptr;


                    obj_ptr = scene->player_obj_ptr;
		    if(obj_ptr != NULL)
		    {
		        type = obj_ptr->type;

		        if(type == SAR_OBJ_TYPE_AIRCRAFT)
		        {
			    sar_object_aircraft_struct *obj_aircraft_ptr =
				obj_ptr->data;
			    if(obj_aircraft_ptr != NULL)
			    {
			        door_ptr = &obj_aircraft_ptr->rescue_door;
			        if(door_ptr->flags & SAR_DOOR_FLAG_STATE)
			        {
				    /* Door is opened, check if hoist is out. */
				    if(obj_aircraft_ptr->hoist.rope_cur > 0)
				    {
				        SARMessageAdd(
					    scene,
				"Cannot close door, hoist is still out."
					);
				        break;
				    }
			        }

		                SARSimOpDoor(
			            scene,
			            &core_ptr->object, &core_ptr->total_objects,
			            obj_ptr,
			            !(door_ptr->flags & SAR_DOOR_FLAG_STATE)
		                );
			    }
		        }
		    }
		}	/* Open/close door. */
	    }
	    break;

	  case 'v': case 'V':
	    if((display != NULL) && state)
            {
		char text[256];

		option.visibility_max += ((display->shift_key_state) ?
		    1 : -1
		);
		if(option.visibility_max > 4)
		    option.visibility_max = 0;
		else if(option.visibility_max < 0)
		    option.visibility_max = 4;

		sprintf(text, "Visibility Max: %i Miles",
		    3 + (option.visibility_max * 3)
		);
                SARMessageAdd(scene, text);
	    }
	    break;

	  case 'c': case 'C':
	    DO_HAS_NO_AUTOREPEAT
            if((display != NULL) && (scene != NULL) && state)
            {
		sar_object_struct *obj_ptr;

		if(!display->ctrl_key_state)
		    break;

		obj_ptr = scene->player_obj_ptr;
		if(obj_ptr->type == SAR_OBJ_TYPE_AIRCRAFT)
		{
		    sar_object_aircraft_struct *obj_aircraft_ptr =
			obj_ptr->data;
		    if(obj_aircraft_ptr != NULL)
		    {
		        obj_aircraft_ptr->engine_state =
			    SAR_ENGINE_STATE_OFF;
			obj_aircraft_ptr->air_worthy_state =
			    SAR_AIR_WORTHY_OUT_OF_CONTROL;

                        SARMessageAdd(scene, "Crash'n ya! :>");
		    }
		}
            }
            break;

          case 'a': case 'A':    /* Tempory function to toggle airplane mode. */
            DO_HAS_NO_AUTOREPEAT
            if((scene != NULL) && state)
            {
                sar_object_struct *obj_ptr;
		sar_object_aircraft_struct *obj_aircraft_ptr;


                obj_ptr = scene->player_obj_ptr;
                if(obj_ptr != NULL)
                {
		    if(obj_ptr->type == SAR_OBJ_TYPE_AIRCRAFT)
		    {
			obj_aircraft_ptr = obj_ptr->data;
/*
			if(obj_aircraft_ptr->engine_can_pitch)
 */
			if(1)
			{
			    switch(obj_aircraft_ptr->flight_model_type)
			    {
			      case SAR_FLIGHT_MODEL_AIRPLANE:
				obj_aircraft_ptr->flight_model_type =
				    SAR_FLIGHT_MODEL_HELICOPTER;
 SARMessageAdd(scene, "Helicopter Mode");
				break;

			      /* All else assume helicopter. */
                              default:
                                obj_aircraft_ptr->flight_model_type =
                                    SAR_FLIGHT_MODEL_AIRPLANE;
 SARMessageAdd(scene, "Airplane Mode");
                                break;
			    }
			}
		    }
                }
            }        
            break;

	  case 'q': case 'Q':	/* Tempory function to move up and down. */
	    DO_HAS_NO_AUTOREPEAT
            if((scene != NULL) && state)
            {
                sar_object_struct *obj_ptr;

                obj_ptr = scene->player_obj_ptr;
                if(obj_ptr != NULL)
                {
		    obj_ptr->pos.z += ((display->shift_key_state) ?
			-150 : 150
		    );
                }
            }
            break;

	  case 'r': case 'R':	/* Tempory function to repair. */
	    DO_HAS_NO_AUTOREPEAT
            if(state)
            {
                sar_object_struct *obj_ptr;
                sar_object_aircraft_struct *obj_aircraft_ptr;
		sar_contact_bounds_struct *cb;

                obj_ptr = scene->player_obj_ptr;
                if(obj_ptr->type == SAR_OBJ_TYPE_AIRCRAFT)
                {
                    obj_aircraft_ptr = obj_ptr->data;
                    if(obj_aircraft_ptr != NULL)
                    {
                        obj_aircraft_ptr->engine_state =
                            SAR_ENGINE_STATE_ON;
                        obj_aircraft_ptr->air_worthy_state =
                            SAR_AIR_WORTHY_FLYABLE;
		    }

		    /* Reset crash flags. */
		    cb = obj_ptr->contact_bounds;
		    if(cb != NULL)
		    {
			cb->crash_flags = SAR_CRASH_FLAG_CRASH_OTHER;
		    }
		}
	    }
	    break;

	  default:
	    /* Some other key, let the game controller handle it. */
	    GCtlHandleKey(
		display,
		core_ptr->gctl,
		c, state,
		display->alt_key_state,
		display->ctrl_key_state,
		display->shift_key_state,
		t
	    );
	    break;
	}

/*
	GWPostRedraw((gw_display_struct *)ptr);
 */

	return;
}
