#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>

#include "../include/os.h"
#ifdef __MSW__
# include <windows.h>
#endif

#include "obj.h"
#include "sound.h"
#include "sar.h"


void SARSceneSoundUpdate(
	sar_core_struct *core_ptr,
	Boolean engine_sounds,
	Boolean event_sounds,
	Boolean voice_sounds,
	Boolean music
);


/*
 *	Updates all sound resources on the core's scene structure and
 *	its objects with respect to the given sound conditions. Purpose
 *	for this function is to start ot stop sounds as the sound options
 *	change in the middle of a game.
 *
 *	Calling this function upon loading a scene or object is not needed
 *	since the loading of scene and objects already loads their sounds
 *	depending on the sound options at the moment the scene and objects
 *	were loaded.
 *
 *	The core structure's scene and recorder must be valid (not NULL).
 *	If the recorder is NULL then it will be initialized.
 */
void SARSceneSoundUpdate(
        sar_core_struct *core_ptr,
        Boolean engine_sounds,
        Boolean event_sounds,
        Boolean voice_sounds,
        Boolean music 
)
{
	int obj_num;
	const char *snd_path;
	sar_scene_struct *scene;
	snd_recorder_struct *recorder;
	sar_object_struct *obj_ptr;
	sar_object_aircraft_struct *obj_aircraft_ptr;
	void **snd_play_rtn;


	if(core_ptr == NULL)
	    return;

	scene = core_ptr->scene;
	recorder = core_ptr->recorder;

	/* Is recorder NULL? */
	if(recorder == NULL)
	{
	    /* Need to initialize recorder. */


/*
	    core_ptr->recorder = recorder;
 */
	    return;
	}

	/* Scene structure must be valid. */
	if(scene == NULL)
	    return;




	/* Update sounds on the scene structure. */




	/* Update sounds on each object. */
	for(obj_num = 0; obj_num < core_ptr->total_objects; obj_num++)
	{
	    obj_ptr = core_ptr->object[obj_num];
	    if(obj_ptr == NULL)
		continue;

	    switch(obj_ptr->type)
	    {
	      case SAR_OBJ_TYPE_AIRCRAFT:
		obj_aircraft_ptr = (sar_object_aircraft_struct *)obj_ptr->data;
		if(obj_aircraft_ptr != NULL)
		{
#define DO_LOAD_ENGINE_SOUND	\
{ \
 if((*snd_play_rtn) == NULL) \
  SARObjLoadSound( \
   recorder, (snd_play_struct **)snd_play_rtn, NULL, \
   snd_path, SND_PLAY_OPTION_REPEATING \
  ); \
}
#define DO_UNLOAD_ENGINE_SOUND	\
{ \
 if((*snd_play_rtn) != NULL) \
 { \
  SoundStopPlay(recorder, *snd_play_rtn); \
  (*snd_play_rtn) = NULL; \
 } \
}

		    if(engine_sounds)
		    {
			/* Turn engine sounds on as needed. */

			snd_play_rtn = &obj_aircraft_ptr->engine_inside_sndplay;
			snd_path = obj_aircraft_ptr->engine_inside_sndpath;
			DO_LOAD_ENGINE_SOUND

                        snd_play_rtn = &obj_aircraft_ptr->engine_outside_sndplay;
                        snd_path = obj_aircraft_ptr->engine_outside_sndpath;
                        DO_LOAD_ENGINE_SOUND
		    }
		    else
		    {
			/* Turn engine sounds off as needed. */

                        snd_play_rtn = &obj_aircraft_ptr->engine_inside_sndplay;
			DO_UNLOAD_ENGINE_SOUND

                        snd_play_rtn = &obj_aircraft_ptr->engine_outside_sndplay;
                        DO_UNLOAD_ENGINE_SOUND
		    }
#undef DO_UNLOAD_ENGINE_SOUND
#undef DO_LOAD_ENGINE_SOUND


		}
		break;

/* Add support for other object types that have sound. */

	    }
	}


	return;
}
